/*************************************************************************
***     Authentication, authorization, accounting + firewalling package
***     (c) 1998-2003 Anton Vinokurov <anton@netams.com>
***		(c) 2003 NeTAMS Development Team
***     	All rights reserved. See 'Copying' file included in distribution
***     	For latest version and more info, visit this project web page
***     	located at http://www.netams.com
***
*** IPFW support added by Dmitry Sergienko <dmitry@trifle.net>.
*** FreeBSD kqueue(2) support added by Roman Nikitchenko <roman@trifle.net>.
*************************************************************************/
/* $Id: ipfw2netflow.c,v 1.2 2003/04/29 11:56:18 jura Exp $ */

#ifdef FREEBSD

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#ifdef KEVENT
#include <sys/event.h>
#endif
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netdb.h>
#include <paths.h>
#include <time.h>
#include <signal.h>
#include <net/ethernet.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif


//---------------------------------------------------------------------------------
/* Cisco NetFlow FlowCollector
 * Copyright (c) 1997 by Cisco Systems, Inc.
 * All rights reserved.
 */

#define FLOW_VERSION_5 5
#define V5FLOWS_PER_PAK 30

/* defaults */
#define EXPIRE_TIMEOUT 5
#define BLOCKSIZE 32768

#define IPV4_HASH(addr1, addr2, port1, port2, proto) ((((addr1) + (addr2)) + (port1) + (port2)+ (proto)) % ((64*1024)))


typedef struct Flow5StatHdr {
    ushort version;
    ushort count;           /* The number of records in PDU. */
    unsigned long SysUptime;       /* Current time in millisecs since router booted */
    unsigned long unix_secs;       /* Current seconds since 0000 UTC 1970 */
    unsigned long unix_nsecs;      /* Residual nanoseconds since 0000 UTC 1970 */
    unsigned long flow_sequence;   /* Seq counter of total flows seen */
    unsigned char engine_type;     /* Type of flow switching engine */
    unsigned char engine_id;       /* ID number of the flow switching engine */
    ushort reserved;
} Flow5StatHdr;


typedef struct IPFlow5Stat {
    struct in_addr srcaddr;    /* Source IP Address */
    struct in_addr dstaddr;    /* Destination IP Address */
    struct in_addr nexthop;    /* Next hop router's IP Address */
    ushort input;          /* Input interface index */
    ushort output;         /* Output interface index */

    unsigned long dPkts;           /* Packets sent in Duration */
    unsigned long dOctets;         /* Octets sent in Duration. */
    unsigned long First;           /* SysUptime at start of flow */
    unsigned long Last;            /* and of last packet of flow */

    unsigned short srcport;        /* TCP/UDP source port number or equivalent */
    unsigned short dstport;        /* TCP/UDP destination port number or equivalent */
    unsigned char pad;
    unsigned char tcp_flags;      /* Cumulative OR of tcp flags */
    unsigned char prot;           /* IP protocol, e.g., 6=TCP, 17=UDP, ... */
    unsigned char tos;            /* IP Type-of-Service */
    unsigned short src_as;         /* originating AS of source address */
    unsigned short dst_as;         /* originating AS of destination address */
    unsigned char src_mask;       /* source address prefix mask bits */
    unsigned char dst_mask;       /* destination address prefix mask bits */
    ushort reserved;
} IPFlow5Stat;

typedef struct {
    Flow5StatHdr header;
    IPFlow5Stat  records[V5FLOWS_PER_PAK];
} IPStat5Msg;

//---------------------------------------------------------------------------------
typedef struct entry {
    u_char ip_tos;
    unsigned long ip_len;
    unsigned long count;
    u_char ip_proto;
    in_addr_t ip_src, ip_dst;
    time_t start;
    time_t last;
    entry *next;
    u_short src_port, dst_port;
    unsigned expired;
    quad_t id;
} entry;

entry **table;
quad_t last_id=0, expired_id=0;
quad_t sent_packets=0, sent_flows=0;
IPStat5Msg *message=NULL;

int quit=0;
int e_count=0;
int debug=0;
int quiet=0;

char *destination=NULL;
char *dst_host=NULL;
int dst_port=20001;
unsigned t_active=600;
unsigned t_inactive=60;
int udp_socket_fd;
struct sockaddr_in udp_dest;
struct timeval tv, tv_now;
#ifdef KEVENT
struct timespec event_ts;
#endif
int alarmcount=1;
int expirecheck_set=0;
int hupset=0;

#define	MAX_PKT_SIZE	0xffff

//#define MY_MALLOC(x) new(x)
#define MY_MALLOC(x) malloc(x)
//#define MY_FREE(x) delete(x)
#define MY_FREE(x) free(x)

void Debug(char *msg,...)
{
#ifdef DEBUG
    va_list args;
    char *params;

    if (!debug || quiet) return;
    params=(char *)malloc(2049);
    va_start(args, msg);
    vsprintf(params, msg, args);
    va_end(args);

    fprintf(stdout, "[%lu]: %s", tv_now.tv_sec, params);
    fflush(stdout);

    free(params);
#else
    return;
#endif
}

void termination(){
    if (debug && !quiet) {
        printf("\nlast_id=%qu, active=%qu\n", last_id, last_id-expired_id);
        printf("sent %qu flows in %qu packets\n", sent_flows, sent_packets);
    }
    quit = 1;
}

void CheckExpire(entry *e) {
    Debug("CheckExpire::begin\n");
    Debug("e->start=%lu, e->last=%lu\n", e->start, e->last);
    if ((unsigned)e->last - (unsigned)e->start > t_active) e->expired=1;
    else if (tv_now.tv_sec - (unsigned)e->last > t_inactive) e->expired=1;
    Debug("%sexpired\n", e->expired?"":"not ");
    Debug("CheckExpire::end\n");
}

unsigned Compare(entry *e, struct ip *ip) {
    struct tcphdr *th;

    if (e->ip_dst!=ip->ip_dst.s_addr) return 0;
    if (e->ip_src!=ip->ip_src.s_addr) return 0;
    if (e->ip_proto!=ip->ip_p) return 0;
    if (e->ip_tos!=ip->ip_tos) return 0;
    if (e->ip_proto==IPPROTO_TCP || e->ip_proto==IPPROTO_UDP) {
        th=(struct tcphdr *)((unsigned char *)ip + ip->ip_hl*4);
        if (e->src_port!=ntohs(th->th_sport)) return 0;
        if (e->dst_port!=ntohs(th->th_dport)) return 0;
    }
    return 1;
}


unsigned Hash(struct ip *ip){
    short sp=0, dp=0;
    struct tcphdr *th;

    if (ip->ip_p==IPPROTO_TCP || ip->ip_p==IPPROTO_UDP) {
        th=(struct tcphdr *)((unsigned char *)ip + ip->ip_hl*4);
        sp=ntohs(th->th_sport);
        dp=ntohs(th->th_dport);
    }
    return IPV4_HASH(ip->ip_src.s_addr, ip->ip_dst.s_addr, sp, dp, ip->ip_p);
}

void Processpacket(struct ip *ip){
    unsigned hash;
    entry *e;
    char buf1[32], buf2[32];
    struct tcphdr *th;

    hash = Hash(ip);
    if (debug && !quiet) {
        strcpy(buf1, inet_ntoa(ip->ip_src));
        strcpy(buf2, inet_ntoa(ip->ip_dst));
    }

    th=(struct tcphdr *)((unsigned char *)ip + ip->ip_hl*4);

    for(e=table[hash]; e!=NULL; e=(entry *)e->next){
        if (Compare(e, ip)) {
            e->ip_len+=ntohs(ip->ip_len);
                e->count++;
                e->last=tv_now.tv_sec;
                break;
        }
    }
    if (e==NULL) { // we must create a new flow record
        e_count++;
        //fprintf(stderr, "malloc %u %u %u\n", __LINE__, sizeof(entry), e_count);
        e=(entry*)malloc(sizeof(entry));
        e->ip_dst=ip->ip_dst.s_addr;
        e->ip_src=ip->ip_src.s_addr;
        e->ip_len=ntohs(ip->ip_len);
        e->ip_proto=ip->ip_p;
        e->ip_tos=ip->ip_tos;
        e->start=e->last=tv_now.tv_sec;
        e->expired=0;
        e->id=last_id;
        e->count=1;
        last_id++;
        e->next=NULL;
        if (e->ip_proto==IPPROTO_TCP || e->ip_proto==IPPROTO_UDP) {
            e->src_port=ntohs(th->th_sport);
            e->dst_port=ntohs(th->th_dport);
        } else {
            e->src_port=ntohs(0);
            e->dst_port=ntohs(0);
        }

        if (!table[hash]) table[hash]=e;
        else {
            entry *t, *tt;
            for (t=table[hash]; t!=NULL; t=(entry *)t->next) tt=t;
            tt->next=e;
        }
    }

    if (debug && !quiet)
        printf("[%d(%qu)]-%p %d s:%s:%u d:%s:%u %d\n", hash, e->id, e, ip->ip_p, buf1, ntohs(th->th_sport), buf2, ntohs(th->th_dport), ntohs(ip->ip_len));
}

void PrintUsage(FILE *f){
    fprintf(f, "Usage: ipfw2netflow {options}\nwhere {options} are:\n\
            -h\t print help screen and exit\n\
            -q\t quiet output\n\
            -d\t debug output (no daemon mode)\n\
            -e export_to\t IP address:port to export flows to\n\
            -b ipfw socket buffer size\n\
            -p ipfw tee/divert bind port\n\
            -x expire check timeout (sec)\n\
            -a active_timeout\t active flow timeout (sec.)\n\
            -i inactive_timeout\t inactive flow timeout (sec.)\n");

}

void DoSend() {
    Debug("DoSend::begin\n");
    if (!message) return;
    sent_packets++;
    sent_flows=sent_flows+message->header.count;
    if (debug && !quiet)
        printf("sending NetFlow v5 packet with %d flows\n", message->header.count);

    message->header.count=ntohs(message->header.count);
    if (sendto(udp_socket_fd, message, sizeof(IPStat5Msg), 0, (struct sockaddr *)&udp_dest, sizeof(udp_dest)) <0)
        printf("UDP send failed[%d]: %s", errno, strerror(errno));

    MY_FREE(message);
    message=NULL;
    Debug("DoSend::end\n");
}

void ToSend(entry *e){
    Debug("ToSend::begin\n");

    if (!message) {
        message=(IPStat5Msg*)MY_MALLOC(sizeof(IPStat5Msg));
        bzero(message, sizeof(IPStat5Msg));
        message->header.version=ntohs(FLOW_VERSION_5);
        message->header.count=0;
        message->header.engine_id=ntohs(0);
        message->header.engine_type=ntohs(31);
        message->header.flow_sequence=ntohl(sent_flows);
    }


    message->records[message->header.count].dOctets=ntohl(e->ip_len);
    message->records[message->header.count].dPkts=ntohl(e->count);
    message->records[message->header.count].input=0;
    message->records[message->header.count].output=0;
    message->records[message->header.count].nexthop.s_addr=INADDR_ANY;

    message->records[message->header.count].tos=e->ip_tos;
    message->records[message->header.count].prot=e->ip_proto;
    message->records[message->header.count].First=e->start;
    message->records[message->header.count].Last=e->last;

    message->records[message->header.count].dst_as=0;
    message->records[message->header.count].dst_mask=0;
    message->records[message->header.count].dstaddr.s_addr=e->ip_dst;
    message->records[message->header.count].dstport=ntohs(e->dst_port);

    message->records[message->header.count].src_as=0;
    message->records[message->header.count].src_mask=0;
    message->records[message->header.count].srcaddr.s_addr=e->ip_src;
    message->records[message->header.count].srcport=ntohs(e->src_port);

    message->header.count++;
    if (debug && !quiet)
        printf("record %d in packet %qu prepared, %ld\n",
               message->header.count-1, sent_packets+1, e->ip_len);
    Debug("Messages in packet: %lu\n", message->header.count);
    if (message->header.count==V5FLOWS_PER_PAK) DoSend();
    Debug("ToSend::end\n");
    
}

void Expiresearch(){
    int i;
    entry *e, *prev;
    time_t t;

    Debug("ExpireSearch::begin\n");

    t=tv_now.tv_sec;
    if (debug && !quiet)
        printf("\nExpireSearch t:[%qu] a:[%qu] at %ld: \n", last_id,last_id-expired_id, t);

    for (i=0; i<65536; i++) {
    JUMP:
        prev=NULL;
        for(e=table[i]; e!=NULL; e=(entry *)e->next) {
            CheckExpire(e);
            if (e->expired) {
                if (debug && !quiet)
                    printf("%qu{%d} ", e->id, i);
                expired_id++;
                if (e==table[i])
                    table[i]=(entry *)e->next;
                else {
                    for(prev=table[i];(prev->next != e) && (prev != NULL);prev=prev->next);
                    if (prev) prev->next=e->next;
                }
                ToSend(e);
                MY_FREE(e);
                e_count--;
                goto JUMP;
            }
            prev=e;
        }
    }

    // it's time to check if there is stalled unsent data in message
    if (message && message->header.count)
        if (message->records[0].Last + 2*t_inactive < (unsigned long) tv_now.tv_sec)
            DoSend();
    Debug("ExpireSearch::end\n");
}

void FlushAll(){
    int i;
    entry *e, *prev;
    time_t t;

    Debug("FlushAll::begin\n");
    t=tv_now.tv_sec;
    if (debug && !quiet)
        printf("\nFlushAll t:[%qu] a:[%qu] at %ld: ", last_id,
               last_id-expired_id, t);

    for (i=0; i<65536; i++) {
    JUMP:
        prev=NULL;
        for(e=table[i]; e!=NULL; e=(entry *)e->next) {
            if (debug && !quiet)
                printf("%qu{%d} ", e->id, i);
            expired_id++;
            if (e==table[i])
                table[i]=(entry *)e->next;
            else {
                for(prev=table[i];(prev->next != e) && (prev != NULL);prev=prev->next);
                if (prev) prev->next=e->next;
            }
            ToSend(e);
            MY_FREE(e);
            e_count--;
            goto JUMP;
        }
        prev=e;
    }

    // it's time to check if there is stalled unsent data in message
    DoSend();
    Debug("FlushAll::begin\n");
}

void do_alarm()
{
    Debug("ALARM, alarmcount=%u\n", alarmcount);
    gettimeofday(&tv_now, 0);
    alarmcount++;
    signal(SIGALRM, (sig_t) do_alarm);
    alarm(1);
    if (alarmcount >= tv.tv_sec) {
        alarmcount = 0;
        expirecheck_set = 1;
    }
}

void do_hup()
{
    Debug("SIGHUP\n");
    hupset = 1;
    signal(SIGHUP, (sig_t)do_hup);
}
//---------------------------------------------------------------------------------

int main(int argc, char **argv)
{
    char 	packet[MAX_PKT_SIZE];
    struct 	sockaddr_in ds;		/* divert socket address */
    socklen_t size_ds = sizeof(ds);
    int 	spkt;			/* divert socket */
#ifdef KEVENT
    struct kevent	event;		/* kevent filter	*/
    int			event_poller;	/* kevent poller	*/
    int			filter_result;	/* main filter result	*/
#else
    fd_set 	allsocks;
#endif
    u_short 	bindport;
    int sock_bsize = BLOCKSIZE;
    int op;

    if (debug && !quiet) {
        printf("ipfw2netflow 1.1 (c) 2002 Anton Vinokurov, anton@netams.com\n");
        printf("This is part of NeTAMS project, http://www.netams.com \n");
    }
    
    bzero(&ds, size_ds);
    ds.sin_family = AF_INET;

    bzero(&tv, sizeof(struct timeval));
    tv.tv_sec = EXPIRE_TIMEOUT;

    table=(entry**)calloc(65536*sizeof(entry*), 1);
    gettimeofday(&tv_now, 0);
    
    while((op=getopt(argc, argv, "hqde:p:a:i:x:b:")) != EOF){
        switch(op){
        case 'h':
            PrintUsage(stdout);
            exit(-1);
            break;
        case 'd':
            debug=1;
            break;
        case 'q':
            quiet=1;
            break;
        case 'e':
            destination=(char *)calloc(strlen(optarg), 1);
            strcpy(destination, optarg);
            break;
        case 'p': /* bind port */
            bindport = atoi(optarg);
            ds.sin_port = htons(bindport);
            break;
        case 'a':
            t_active=atoi(optarg);
            break;
        case 'i':
            t_inactive=atoi(optarg);
            break;
        case 'x':
            tv.tv_sec=atoi(optarg);
            break;
        case 'b':
            sock_bsize=atoi(optarg);
            break;
        default:
            PrintUsage(stdout);
            break;
        }
    }

    if (!ds.sin_port) {
        printf("Please define tee port with -p flag!\n\n");
        PrintUsage(stdout);
        exit(1);
    }
    if (!destination)
        dst_host="127.0.0.1";
    else {
        char *c;
        c=strchr(destination, ':');
        if (c) {
            dst_port=atoi(c+1);
            c[0]='\0'; dst_host=destination;
        } else
            dst_host=destination;
    }
    
    bzero(&udp_dest, sizeof(udp_dest));
    udp_dest.sin_family=AF_INET;
    udp_dest.sin_port=htons(dst_port);
    inet_aton(dst_host, &udp_dest.sin_addr);
    udp_socket_fd=socket(AF_INET, SOCK_DGRAM, 0);
    if (udp_socket_fd==-1)
    {
        if (!quiet) fprintf(stderr, "failed to create UDP socket: %s\n", strerror(errno));
        exit(5);
    }

    if ((spkt = socket(PF_INET, SOCK_RAW, IPPROTO_DIVERT)) == -1)
        fprintf(stderr, "socket(IP_DIVERT)");

    if ((bind(spkt, (struct sockaddr *)&ds, sizeof(ds))) != 0)
        fprintf(stderr, "bind");

    if (!debug)
    {
        daemon(0, 1);
        chdir("/var/tmp");
        umask(077);
        fclose(stdin);
        fclose(stdout);
        fclose(stderr);
    }

#ifdef KEVENT
    event_poller = kqueue();
    if (event_poller < 0)
    {
        if (!quiet)
	    fprintf(stderr, "failed to create kqueue: %s\n", strerror(errno));
        exit(5);
    }

    bzero( &event, sizeof( struct kevent ) );
    bzero( &event_ts, sizeof( struct timespec ) );

    event.ident  = spkt;
    event.filter = EVFILT_READ;
    event.flags  = EV_ADD | EV_ENABLE;

    if ( kevent( event_poller, &event, 1, NULL, 0, &event_ts ) < 0 )
    {
        if (!quiet)
	    fprintf(stderr, "kevent filter failed: %s\n", strerror(errno));
        exit(5);
    }
    event_ts.tv_sec = tv.tv_sec;
#else
    FD_ZERO(&allsocks);
    FD_SET(spkt, &allsocks);
#endif

    if (sock_bsize)
    {
        setsockopt(spkt, SOL_SOCKET, SO_RCVBUF, &sock_bsize, sizeof(int));
        setsockopt(spkt, SOL_SOCKET, SO_SNDBUF, &sock_bsize, sizeof(int));
    }

    signal(SIGINT, (sig_t)termination);
    signal(SIGQUIT, (sig_t)termination);
    signal(SIGTERM, (sig_t)termination);
    signal(SIGHUP, (sig_t) do_hup);
    signal(SIGALRM, (sig_t) do_alarm);
    alarm(1);

    while(!quit)
    {
        long len;

#ifdef KEVENT
	struct kevent result;

        if ( (filter_result =
	      kevent( event_poller, NULL, 0, &result, 1, NULL )) < 0 )
#else
        fd_set readable;

        readable = allsocks;
        if (select(spkt + 1, &readable, NULL, NULL, &tv) < 0)
#endif
	{
            if (errno != EINTR)
	    {
                quit = 1;
                break;
            }
            Debug("alarmcount=%u, expirecheck_set=%u\n", alarmcount, expirecheck_set);
            if (expirecheck_set == 1) {
                Expiresearch();
                expirecheck_set=0;
            }

            if (hupset)
            {
                FlushAll();
                hupset=0;
            }
            continue;
        }


        /* check for packet available */
#ifdef KEVENT
	if ( (filter_result == 1) && (((int)result.ident) == spkt) )
#else
        if (FD_ISSET(spkt, &readable))
#endif
	{
            if ((len = recvfrom(spkt, packet, MAX_PKT_SIZE, 0,
                                (struct sockaddr *)&ds, &size_ds)) != -1)
	    {
                Processpacket((struct ip *)packet);
	    }
        }
    }

    FlushAll();
#ifdef KEVENT
    close( event_poller );
#else
    FD_CLR(spkt, &allsocks);
#endif
    close(spkt);
    close(udp_socket_fd);

    MY_FREE(table);

    signal(SIGINT, SIG_DFL);
    signal(SIGQUIT, SIG_DFL);
    signal(SIGTERM, SIG_DFL);
    signal(SIGALRM, SIG_DFL);
    signal(SIGHUP, SIG_DFL);
#ifdef DMALLOC    
    dmalloc_shutdown();
#endif    
    return 0;
}

#else

#include <stdio.h>

int main() {
        printf("ipfw2netflow 1.1 (c) 2003 Anton Vinokurov, anton@netams.com\n");
        printf("This is part of NeTAMS project, http://www.netams.com \n");
	printf("This software could be started under FreeBSD operating system only\n");
}

#endif // FREEBSD
